/*

    bag -- A quick hack to buffer audio data between TiMidity and an
           external player program.

    This program is in the Public Domain. It is distributed without
    much hope that it will be useful, and WITHOUT ANY WARRANTY;
    definitely without any implied warranty of MERCHANTABILITY or
    FITNESS FOR ANY PURPOSE.

    bag may enable you to play complex files at higher output rates,
    but it's a bit of a pain to use.

    Here's a usage example for Linux. splay is a program that (with
    this invocation) plays audio data from stdin. Your flag
    requirements may vary.

		timidity -id -Or1lsS -o- -s22 -p48 filenames |
		bag -b 2M -t 1M -c 4k |
		splay -b16 -s22 -S

*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <time.h>
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

static int bytes(char *s)
{
  int i;
  i=atoi(s);
  while (*s>='0' && *s<='9')
    s++;

  switch(*s)
    {
    case 'm':
    case 'M':
      i *= 1024; /* note: this will fall-through and do *1024 again */
    case 'k':
    case 'K':
      i *= 1024;
    }
  return i;
}

static char s_junk[20], s_tw[20], s_tr[20];

#define num2str(x,y) sprintf(buf, "%5.1f%c",x,y)
int unbytes(char *buf, long long num)
{
  float f=(float) num;
  
  if (num < 9999) return sprintf(buf, "%4d", num);
  f=f/1024; if (f < 9999.0) return num2str(f,'K');
  f=f/1024; if (f < 9999.0) return num2str(f,'M');
  f=f/1024; if (f < 9999.0) return num2str(f,'G');
  f=f/1024; if (f < 9999.0) return num2str(f,'T');
  f=f/1024; if (f < 9999.0) return num2str(f,'E');
  
  strcpy (buf, "+inf.");
  return 0;
}

int main(int argc, char **argv)
{
  int bs=1024*1024, rs=4*1024, ws=4*1024, cs=4*1024, threshold, thresh=64*1024;
  int c, c0, c1, junk=0, verbosity=0, flush_and_go=0, tsec=1, hwmark, cnt, maxcnt=5;
  long long int totalr=0, totalw=0;
  char *bag, *head, *tail;
  extern char *optarg; /* for Sun's gcc */
  time_t lasttime=0;

  fd_set readfds, writefds, brokefds;
  struct timeval tmo;

  while ((c=getopt(argc, argv, "b:t:r:w:c:hvm:o:"))>0)
    switch(c)
      {
      case 'b': bs=bytes(optarg); break;
      case 't': thresh=bytes(optarg); break;
      case 'r': rs=bytes(optarg); break;
      case 'w': ws=bytes(optarg); break;
      case 'c': rs=ws=bytes(optarg); break;
      case 'm': tsec=atoi(optarg); break;
      case 'n': maxcnt=atoi(optarg); break;
      case 'v': verbosity++; break;
      case 'o': {
         c=open(optarg, O_WRONLY);
	 if (c<0) perror ("open out");
	 close (STDOUT_FILENO);
	 if (!dup2(c,STDOUT_FILENO)) perror ("dup2");
	 close (c);
	 fprintf(stderr, "opened %s for output\n", optarg);
	 break;
      }
      case '?':
      case 'h':
	fprintf(stderr, "\nusage: bag [options]\n\n"
	       "  -b size    Set buffer size (default %d)\n"
	       "  -r size    Set read chunk size (default %d)\n"
	       "  -w size    Set write chunk size (default %d)\n"
	       "  -c size    Set read/write chunk size\n"
	       "  -t size    Set start threshold for writing (default %d)\n"
	       "  -m sec     seconds to wait between R/W (default %d)\n"
	       "  -o fname   output file if not stdout\n"
	       "  -n num     after num of buffer underruns, resets threshold (default %d)\n"
	       "  -v         Verbose mode (more for more)\n"
	       "  -h         This help screen\n\n"
	       ,bs,rs,ws,thresh,tsec,maxcnt);
	return 0;
      }

  if (ws>rs) cs=ws;
  else cs=rs;
  if (bs < ws+rs)
    {
      bs=ws+rs;
      fprintf(stderr, "bag: Buffer size increased to %d bytes\n", bs);
    }
  if (thresh > (bs-rs))
    {
      thresh=bs-rs;
      fprintf(stderr, "bag: Threshold lowered to %d bytes\n", bs);
    }

  hwmark=bs-2*ws;
  if (hwmark<=ws) hwmark=bs; /* force blocking if buff too small */
  
  if (!(head=tail=bag=malloc(bs))) perror("bag: malloc failed! Zaphf!");

  junk=0;
  cnt=maxcnt;
  threshold=thresh;
  
  while (!flush_and_go)
    {
      FD_ZERO(&readfds);
      FD_ZERO(&writefds);
      FD_ZERO(&brokefds);

      if (junk<(bs-rs))
	FD_SET(STDIN_FILENO, &readfds);

      if (junk>ws && junk>thresh)
	{
	  FD_SET(STDOUT_FILENO, &writefds);
	  thresh=0;
	}
      
      FD_SET(STDIN_FILENO, &brokefds);
      FD_SET(STDOUT_FILENO, &brokefds);
      
      tmo.tv_sec=tsec;
      tmo.tv_usec=0;
#ifdef HPUX
      if ((c=select(FD_SETSIZE,(int *) &readfds,(int *) &writefds, (int *) &brokefds, &tmo))<0)
#else
      if ((c=select(FD_SETSIZE, &readfds, &writefds, &brokefds, &tmo))<0)
#endif
	perror("\nbag: select");


      if (FD_ISSET(STDOUT_FILENO, &brokefds))
	{
	  fprintf(stderr, " Error writing output stream!");
	  /*return 1;*/
	}
	
      c1=0; /* how much we red ? */
      if (FD_ISSET(STDIN_FILENO, &readfds))
	{
	  if ((c1=c=read(STDIN_FILENO, head, 
		      (head+rs >= bag+bs) ? (bag+bs-head) : rs))<=0)
	    {
	      if (c==0 || errno!=EAGAIN)
		flush_and_go=1;
	      if (c<0) perror("\nbag: I died like a rebel scum at read!");
	    }
	  else
	    {
	      junk+=c;
	      head+=c;
	      if (head>=bag+bs)
		head-=bs;
	      totalr+=c;
	    }
	}
      else if (FD_ISSET(STDIN_FILENO, &brokefds))
	flush_and_go=1;


/*
 * if read has  terminated (flush_and_go) or
 * if we are free to write (FD_ISSET) or
 * if we above high watermark or
 * last read did not read anything anyway (it's stalling) (and we have at least on write block to write)
 * THEN we will write one block of data...
 */
 
      if ((flush_and_go || FD_ISSET(STDOUT_FILENO, &writefds) || junk > hwmark || (!c1 && junk >=ws)) && thresh == 0 )
	do
	  {
	    c=(tail+ws >= bag+bs) ? (bag+bs-tail) : ws;
	    if (c>junk) c=junk;
	    
	    if (flush_and_go && verbosity)
	      fprintf(stderr, "\rbag: Flushing, %8d bytes", junk);

	    c0=c;
	    c=write(STDOUT_FILENO, tail, c);
	    if (c < 0) perror("\nbag: I died at write!");
	    if (c != c0) {
	      fprintf(stderr, "\nbag: short write! (%d != %d)", c, c0);
	      /*exit(2);*/
	    }
	    junk -= c;
	    tail += c;
            totalw += c;
	    if (tail>=bag+bs)
	      tail-=bs;
	  } while (flush_and_go && junk);

       if (verbosity && time(NULL)-lasttime>1) {
         unbytes ((char *)&s_junk, junk);
         unbytes ((char *)&s_tr, totalr);
         unbytes ((char *)&s_tw, totalw);
	 fprintf(stderr, "\rbag: %s bytes [total read/write: %s/%s]    ", (char *)&s_junk, (char *)&s_tr, (char *)&s_tw);
	 lasttime = time(NULL);
       }
       if (junk < 2*rs) --cnt;
       if (cnt < 0 && thresh == 0) {
         fprintf(stderr," reseting threshold, throttling...\n");
         thresh = threshold;
	 cnt = maxcnt;
       }

    }
  if (verbosity)
    fprintf(stderr, "\nbag: All done\n");
  return 0;
}
